import os
import numpy as np
from .bert_score import score
BERTSCORE_BATCH = int(os.getenv('BERTSCORE_BATCH', '32'))


class BERTScore:

    def __init__(self, language: str = 'en'):
        self.language = language

    def get_score(self, hyps, refs):
        return np.array(score(hyps, refs, lang=self.language, verbose=True, batch_size=BERTSCORE_BATCH)[0].tolist())

    def compute_score(self, gts, res):
        assert gts.keys() == res.keys()
        _ids = gts.keys()
        hyps = [res[_id][0].decode() for _id in _ids]
        refs = [gts[_id][0].decode() for _id in _ids]
        _score = self.get_score(hyps, refs)
        return np.mean(_score), _score

    @staticmethod
    def method():
        return "BERTScore"



import logging
from typing import List
from itertools import product, chain
from statistics import mean
import numpy as np
from lmqg.spacy_module import SpacyPipeline

EPS = 1e-6


def get_score(base_metric, language: str = 'en'):
    if base_metric == 'bertscore':
        from .bertscore import BERTScore
        return BERTScore(language=language)
    if base_metric == 'moverscore':
        from .moverscore import MoverScore
        return MoverScore(language=language)
    raise ValueError(f'unknown metric {base_metric}: metric should be `moverscore`/`bertscore`')


class QAAlignedF1Score:

    def __init__(self,
                 language: str = 'en',
                 target_metric: str = 'f1',
                 base_metric: str = 'bertscore',
                 instance_separator: str = " | ",
                 question_key: str = 'question: ',
                 answer_key: str = 'answer: ',
                 qa_separator: str = ', '):
        self.language = language
        self.base_metric = get_score(base_metric, language)
        self.instance_separator = instance_separator
        self.question_key = question_key
        self.answer_key = answer_key
        # pipe = SpacyPipeline(self.language)
        # self.question_key = (pipe.token(question_key))
        # self.answer_key = pipe.token(answer_key)
        self.qa_separator = qa_separator
        assert target_metric in ['f1', 'recall', 'precision'], target_metric
        self.target_metric = target_metric

    def sanity_check(self, sample: str):
        if len(sample.split(self.qa_separator + self.answer_key)) != 2:
            logging.info(f'error (length != 2): {sample}')
            return False
        if not sample.startswith(self.question_key):
            logging.info(f'error (question not found): {sample}')
            return False
        return True

    def filter_qa_pairs(self, qa_pairs):
        qa_pairs = list(set(qa_pairs))  # remove duplication
        qa_pairs = filter(lambda x: self.sanity_check(x), qa_pairs)
        return list(qa_pairs)

    def get_score(self, hyps: List, refs: List):
        hyps = [h.split(self.instance_separator) for h in hyps]
        refs = [r.split(self.instance_separator) for r in refs]
        hyps = [self.filter_qa_pairs(hyp) for hyp in hyps]
        logging.info(f"found {len([i for i in hyps if len(i) == 0])} empty prediction from {len(hyps)}")
        pairs = list(chain(*[list(product(h, r)) for h, r in zip(hyps, refs) if len(h) != 0]))
        if len(pairs) == 0:
            return np.array([0.0])
        h, r = list(zip(*pairs))
        scores = self.base_metric.get_score(h, r)
        assert len(scores) == len(pairs), f"{len(scores)} != {len(pairs)}"
        pair_score = {f"{h}--{r}": s for (h, r), s in zip(pairs, scores)}
        output = []
        for hyp, ref in zip(hyps, refs):
            if len(hyp) == 0:
                output.append({"f1": 0, "precision": 0, "recall": 0})
            else:
                precision = mean(max(pair_score[f"{h}--{r}"] for r in ref) for h in hyp)
                recall = mean(max(pair_score[f"{h}--{r}"] for h in hyp) for r in ref)
                f1 = 2 * precision * recall / (precision + recall + EPS)
                output.append({"f1": f1, "precision": precision, "recall": recall})
        output = np.array([o[self.target_metric] for o in output])
        return output

    def compute_score(self, gts, res):
        assert gts.keys() == res.keys()
        _ids = gts.keys()
        hyps = [res[_id][0].decode() for _id in _ids]
        refs = [gts[_id][0].decode() for _id in _ids]
        _score = self.get_score(hyps, refs)
        return np.mean(_score), _score

    @staticmethod
    def method():
        return "QAAlignedF1Score"



""" Evaluate Multitask QAG Model with QAG metric (eg. QA Aligned F1). """
import json
import logging
import argparse
import os

from evaluate import load
from lmqg import TransformersQG, get_dataset


squad_metric = load("squad")
logging.basicConfig(format='%(asctime)s %(levelname)-8s %(message)s', level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')


def get_options():
    parser = argparse.ArgumentParser(description='QA evaluation (F1 & Exact Match)')
    parser.add_argument('-m', '--model-checkpoint', default=None, type=str)
    parser.add_argument('--max-length', default=512, type=int, help='')
    parser.add_argument('--max-length-output', default=256, type=int, help='')
    parser.add_argument('-d', '--dataset-path', help='huggingface datasets alias', default='lmqg/qg_squad', type=str)
    parser.add_argument('--dataset-name', help='huggingface datasets name', default='default', type=str)
    parser.add_argument('--test-split', help='the name of test split', default='test', type=str)
    parser.add_argument('--validation-split', help='the name of validation split', default='validation', type=str)
    parser.add_argument('--n-beams', default=4, type=int)
    parser.add_argument('--batch-size', default=16, type=int)
    parser.add_argument('-l', '--language', help='', default='en', type=str)
    parser.add_argument('--use-auth-token', help='', action='store_true')
    parser.add_argument('-e', '--export-dir', required=True, type=str)
    parser.add_argument('--hyp-test', default=None, type=str)
    parser.add_argument('--hyp-dev', default=None, type=str)
    parser.add_argument('--overwrite-prediction', help='', action='store_true')
    parser.add_argument('--overwrite-metric', help='', action='store_true')
    parser.add_argument('-i', '--input-type', help='', default='paragraph_question', type=str)
    parser.add_argument('-o', '--output-type', help='', default='answer', type=str)
    return parser.parse_args()


def main():

    opt = get_options()
    os.makedirs(opt.export_dir, exist_ok=True)

    def load_model():
        if opt.model_checkpoint is not None:
            _model = TransformersQG(opt.model_checkpoint,
                                    skip_overflow_error=True,
                                    drop_answer_error_text=True,
                                    language=opt.language,
                                    max_length=opt.max_length,
                                    max_length_output=opt.max_length_output,
                                    use_auth_token=opt.use_auth_token)
            _model.eval()
            return _model
        raise ValueError(f"require `-m` or `--model-checkpoint`")

    metric_file = f"{opt.export_dir}/metric.first.answer.{opt.input_type}.{opt.output_type}." \
                  f"{opt.dataset_path.replace('/', '_')}.{opt.dataset_name}.json"
    if os.path.exists(metric_file):
        with open(metric_file) as f:
            output = json.load(f)
    else:
        output = {}

    for _split, _file in zip([opt.test_split, opt.validation_split], [opt.hyp_test, opt.hyp_dev]):
        _file = f"{opt.export_dir}/samples.{_split}.hyp.{opt.input_type}.{opt.output_type}." \
                f"{opt.dataset_path.replace('/', '_')}.{opt.dataset_name}.txt" if _file is None else _file
        logging.info(f'generate qa for split {_split}')
        if _split not in output:
            output[_split] = {}

        data_in, gold_reference = get_dataset(
            opt.dataset_path,
            opt.dataset_name,
            input_type=opt.input_type,
            output_type=opt.output_type,
            split=_split,
            use_auth_token=opt.use_auth_token)
        references = [{"answers": {"answer_start": [100], "text": [r]}, "id": str(n)} for n, r in
                      enumerate(gold_reference)]
        prediction = None
        if not opt.overwrite_prediction and os.path.exists(_file):
            with open(_file) as f:
                _prediction = f.read().split('\n')
            if len(_prediction) != len(gold_reference):
                logging.warning(f"found prediction file at {_file} but length not match "
                                f"({len(_prediction)} != {len(gold_reference)})")
            else:
                prediction = _prediction
        if prediction is None:
            model = load_model()
            prediction = model.generate_prediction(
                inputs=data_in,
                prefix_type="qa" if model.add_prefix else None,
                num_beams=opt.n_beams,
                batch_size=opt.batch_size)
            # formatting prediction
            assert len(prediction) == len(data_in), f"{len(prediction)} != {len(data_in)}"
            with open(_file, 'w') as f:
                f.write('\n'.join(prediction))

        prediction = [{"prediction_text": p, "id": str(n)} for n, p in enumerate(prediction)]

        if opt.overwrite_metric or 'F1 (QA)' not in output[_split] or 'Exact Match (QA)' not in output[_split]:
            scores = squad_metric.compute(predictions=prediction, references=references)
            output[_split]['AnswerF1Score'] = scores['f1']
            output[_split]['AnswerExactMatch'] = scores['exact_match']

    with open(metric_file, "w") as f:
        json.dump(output, f)




""" Custom exceptions """


class ExceedMaxLengthError(Exception):
    """ Token exceed max length. """

    def __init__(self, max_length=None):
        self.message = f'Input sentence exceeds max length of {max_length}'
        super().__init__(self.message)


class HighlightNotFoundError(Exception):
    """ Highlight is not in the sentence. """

    def __init__(self, highlight: str, input_sentence: str):
        self.message = f'Highlight `{highlight}` not found in the input sentence `{input_sentence}`'
        super().__init__(self.message)


class AnswerNotFoundError(Exception):
    """ Answer cannot found in the context. """

    def __init__(self, context: str):
        self.message = f'Model cannot find any answer candidates in `{context}`'
        super().__init__(self.message)


class APIError(Exception):
    """ Error from huggingface inference API. """

    def __init__(self, context: str):
        self.message = f'Huggingface API Error:\n`{context}`'
        super().__init__(self.message)



"""https://github.com/AIPHES/emnlp19-moverscore/blob/b9d4144e785293849216618e85db887b3ac741d3/moverscore_v2.py"""
import os
import string
from collections import defaultdict
from typing import List
from tqdm import tqdm

import torch
import numpy as np
from pyemd import emd_with_flow
from transformers import AutoTokenizer, AutoModel

MOVERSCORE_BATCH = int(os.getenv('MOVERSCORE_BATCH', '32'))
# lang2model = defaultdict(lambda: "bert-base-multilingual-cased")
# lang2model.update(
#     {
#         "en": "roberta-large",
#         "zh": "bert-base-chinese",
#         "tr": "dbmdz/bert-base-turkish-cased",
#         "en-sci": "allenai/scibert_scivocab_uncased",
#     }
# )

lang2model = defaultdict(lambda: "bert-base-multilingual-cased")
lang2model.update({"en": "distilbert-base-uncased"})


def batched_cdist_l2(x1, x2):
    x1_norm = x1.pow(2).sum(dim=-1, keepdim=True)
    x2_norm = x2.pow(2).sum(dim=-1, keepdim=True)
    res = torch.baddbmm(x2_norm.transpose(-2, -1), x1, x2.transpose(-2, -1), alpha=-2).add_(
        x1_norm).clamp_min_(1e-30).sqrt_()
    return res


class MoverScore:

    def __init__(self, language: str = 'en'):
        model_name = lang2model[language]
        try:
            self.tokenizer = AutoTokenizer.from_pretrained(model_name, do_lower_case=True)
        except ValueError:
            self.tokenizer = AutoTokenizer.from_pretrained(model_name, do_lower_case=True, local_files_only=True)
        try:
            self.model = AutoModel.from_pretrained(model_name, output_hidden_states=True, output_attentions=True)
        except ValueError:
            self.model = AutoModel.from_pretrained(model_name, output_hidden_states=True, output_attentions=True,
                                                   local_files_only=True)
        self.model_name = model_name
        self.device = 'cuda' if torch.cuda.device_count() > 0 else 'cpu'
        self.model.to(self.device)
        self.model.eval()

    def collate_idf(self, arr, idf_dict, pad="[PAD]"):

        def padding(_arr, dtype=torch.long):
            lens = torch.LongTensor([len(a) for a in _arr])
            max_len = lens.max().item()
            padded = torch.ones(len(_arr), max_len, dtype=dtype) * self.tokenizer.convert_tokens_to_ids([pad])[0]
            mask = torch.zeros(len(_arr), max_len, dtype=torch.long)
            for i, a in enumerate(_arr):
                padded[i, :lens[i]] = torch.tensor(a, dtype=dtype)
                mask[i, :lens[i]] = 1
            return padded, lens, mask

        def truncate(_tokens):
            if len(_tokens) > self.tokenizer.model_max_length - 2:
                return _tokens[0:(self.tokenizer.model_max_length - 2)]
            return _tokens

        tokens = [["[CLS]"] + truncate(self.tokenizer.tokenize(a)) + ["[SEP]"] for a in arr]
        arr = [self.tokenizer.convert_tokens_to_ids(a) for a in tokens]
        idf_weights = [[idf_dict[i] for i in a] for a in arr]
        _padded, _lens, _mask = padding(arr, dtype=torch.long)
        padded_idf, _, _ = padding(idf_weights, dtype=torch.float)

        return _padded, padded_idf, _lens, _mask, tokens

    def get_bert_embedding(self, all_sens, idf_dict, batch_size=-1):

        def bert_encode(x, attention_mask):
            result = self.model(x.to(self.device), attention_mask=attention_mask.to(self.device))
            return result[1] if self.model_name == 'distilbert-base-uncased' else result[2]

        padded_sens, padded_idf, lens, mask, tokens = self.collate_idf(all_sens, idf_dict)

        batch_size = len(all_sens) if batch_size == -1 else batch_size
        embeddings = []
        with torch.no_grad():
            for i in range(0, len(all_sens), batch_size):
                batch_embedding = bert_encode(padded_sens[i:i + batch_size], attention_mask=mask[i:i + batch_size])
                batch_embedding = torch.stack(batch_embedding).cpu()
                embeddings.append(batch_embedding)
                del batch_embedding

        total_embedding = torch.cat(embeddings, dim=-3)
        return total_embedding, lens, mask, padded_idf, tokens

    def word_mover_score(self, refs, hyps, idf_dict_ref, idf_dict_hyp, stop_words=None, batch_size=256):
        stop_words = [] if stop_words is None else stop_words
        preds = []
        for batch_start in tqdm(list(range(0, len(refs), batch_size))):
            batch_refs = refs[batch_start:batch_start + batch_size]
            batch_hyps = hyps[batch_start:batch_start + batch_size]

            ref_embedding, ref_lens, ref_masks, ref_idf, ref_tokens = self.get_bert_embedding(batch_refs, idf_dict_ref)
            hyp_embedding, hyp_lens, hyp_masks, hyp_idf, hyp_tokens = self.get_bert_embedding(batch_hyps, idf_dict_hyp)

            ref_embedding = ref_embedding[-1]
            hyp_embedding = hyp_embedding[-1]

            batch_size = len(ref_tokens)
            for i in range(batch_size):
                ref_ids = [k for k, w in enumerate(ref_tokens[i]) if
                           w in stop_words or '##' in w or w in set(string.punctuation)]
                hyp_ids = [k for k, w in enumerate(hyp_tokens[i]) if
                           w in stop_words or '##' in w or w in set(string.punctuation)]

                ref_embedding[i, ref_ids, :] = 0
                hyp_embedding[i, hyp_ids, :] = 0

                ref_idf[i, ref_ids] = 0
                hyp_idf[i, hyp_ids] = 0

            raw = torch.cat([ref_embedding, hyp_embedding], 1)
            raw.div_(torch.norm(raw, dim=-1).unsqueeze(-1) + 1e-30)
            distance_matrix = batched_cdist_l2(raw, raw).double().cpu().numpy()

            for i in range(batch_size):
                c1 = np.zeros(raw.shape[1], dtype=float)
                c2 = np.zeros(raw.shape[1], dtype=float)
                c1[:len(ref_idf[i])] = ref_idf[i]
                c2[len(ref_idf[i]):] = hyp_idf[i]
                c1 = c1 / (np.sum(c1) + 1e-30)
                c2 = c2 / (np.sum(c2) + 1e-30)
                dst = distance_matrix[i]
                _, flow = emd_with_flow(c1, c2, dst)
                flow = np.array(flow, dtype=np.float32)
                score = 1. / (1. + np.sum(flow * dst))  # 1 - np.sum(flow * dst)
                preds.append(score)

        return preds

    def get_score(self, hyps, refs):
        assert len(hyps) == len(refs)
        idf_dict_hyp = defaultdict(lambda: 1.)
        idf_dict_ref = defaultdict(lambda: 1.)
        return np.array(self.word_mover_score(refs, hyps, idf_dict_ref, idf_dict_hyp, batch_size=MOVERSCORE_BATCH))

    def compute_score(self, gts, res):
        assert gts.keys() == res.keys()
        _ids = gts.keys()
        hyps = [res[_id][0].decode() for _id in _ids]
        refs = [gts[_id][0].decode() for _id in _ids]
        _score = self.get_score(hyps, refs)
        return np.mean(_score), _score

    @staticmethod
    def method():
        return "MoverScore"



""" Data utility """
import os
import requests
from os.path import join as pj

from datasets import load_dataset
from .language_model import internet_connection


__all__ = ('get_dataset', 'get_reference_files', 'DEFAULT_CACHE_DIR')

DEFAULT_CACHE_DIR = pj(os.path.expanduser('~'), '.cache', 'lmqg')
# dataset requires custom reference file
DATA_NEED_CUSTOM_REFERENCE = ['lmqg/qg_squad']


def get_dataset(path: str = 'lmqg/qg_squad',
                name: str = 'default',
                split: str = 'train',
                input_type: str = 'paragraph_answer',
                output_type: str = 'question',
                use_auth_token: bool = False):
    """ Get question generation input/output list of texts. """
    name = None if name == 'default' else name
    dataset = load_dataset(path, name, split=split, use_auth_token=use_auth_token)
    return dataset[input_type], dataset[output_type]


def get_reference_files(path: str = 'lmqg/qg_squad', name: str = 'default', cache_dir: str = None):
    """ Get reference files for automatic evaluation """
    name = None if name == 'default' else name
    local_files_only = not internet_connection()
    cache_dir = pj(DEFAULT_CACHE_DIR, 'reference_files', path) if cache_dir is None else cache_dir
    output = {}
    url = f'https://huggingface.co/datasets/{path}/raw/main/reference_files'
    for split in ['test', 'validation']:
        dataset = load_dataset(path, name, split=split)
        for feature in ['answer', 'question', 'paragraph', 'sentence', 'questions_answers']:
            if feature not in dataset.features:
                continue
            filename = f'{feature}-{split}.txt' if name is None else f'{feature}-{split}.{name}.txt'
            ref_path = pj(cache_dir, filename)
            os.makedirs(os.path.dirname(ref_path), exist_ok=True)
            if path in DATA_NEED_CUSTOM_REFERENCE:
                if not os.path.exists(ref_path):
                    assert not local_files_only, f'network is not reachable, could not download the file from {url}/{filename}'
                    r = requests.get(f'{url}/{filename}')
                    content = r.content
                    assert "Entry not found" not in str(content) and content != b'', content
                    with open(ref_path, "wb") as f:
                        f.write(content)
                    with open(ref_path) as f:
                        assert len(f.read().split('\n')) > 20, f"invalid file {ref_path}"
            else:
                with open(ref_path, 'w') as f:
                    f.write('\n'.join([i.replace('\n', '.') for i in dataset[feature]]))
            assert os.path.exists(ref_path)
            output[f'{feature}-{split}'] = ref_path
    return output



from .grid_searcher import GridSearcher
from .trainer import Trainer
from .language_model import TransformersQG
from .data import get_dataset, get_reference_files, DEFAULT_CACHE_DIR
from .automatic_evaluation import evaluate, compute_metrics



""" Spacy Pipeline
python -m spacy download en_core_web_sm  # English
python -m spacy download ja_core_news_sm  # Japanese
python -m spacy download zh_core_web_sm  # Chinese
python -m spacy download de_core_news_sm  # German
python -m spacy download es_core_news_sm  # Spanish
python -m spacy download it_core_news_sm  # Italy
python -m spacy download ko_core_news_sm  # Korean
python -m spacy download ru_core_news_sm  # Russian
python -m spacy download fr_core_news_sm  # French
"""
import spacy

__all__ = 'SpacyPipeline'

MODELS = {
    "en": "en_core_web_sm",
    "ja": "ja_core_news_sm",
    "zh": "zh_core_web_sm",
    "de": "de_core_news_sm",
    "es": "es_core_news_sm",
    "it": "it_core_news_sm",
    "ko": "ko_core_news_sm",
    "ru": "ru_core_news_sm",
    "fr": "fr_core_news_sm"
}
VALID_METHODS = ['positionrank', 'textrank', 'biasedtextrank', 'positionrank', 'ner']


class SpacyPipeline:

    def __init__(self, language, algorithm: str = None):
        model = "en_core_web_sm" if language not in MODELS else MODELS[language]

        self.nlp = spacy.load(model)
        self.nlp.add_pipe("sentencizer")
        self.algorithm = algorithm
        self.library = None
        if self.algorithm is not None and self.algorithm != 'ner':
            assert algorithm in VALID_METHODS, f'invalid algorithm {algorithm}\n- valid list: {VALID_METHODS}'
            if self.algorithm == 'yake':
                import spacy_ke  # need to load yake
                self.nlp.add_pipe("yake")
                self.library = 'spacy_ke'
            elif self.algorithm in ['textrank', 'biasedtextrank', 'positionrank']:
                import pytextrank
                self.nlp.add_pipe(algorithm)
                self.library = 'pytextrank'
            else:
                raise ValueError(f'unknown algorithm: {self.algorithm}')

    def _get_keyword(self, output, original_document=None, n=None):
        if self.algorithm == 'ner':
            return [str(i) for i in output.ents]
        assert original_document is not None
        assert n is not None
        if self.library == 'spacy_ke':
            return [str(term) for term, score in output._.extract_keywords(n) if str(term) in original_document]
        return [str(i.text) for i in output._.phrases[:n] if str(i.text) in original_document]

    def sentence_keyword(self, string: str, n: int = 10):
        out = self.nlp(string)
        sentence = [str(i) for i in out.sents if len(i) > 0]
        keyword = self._get_keyword(out, string, n)
        return sentence, keyword

    def sentence(self, string: str):
        return [str(i) for i in self.nlp(string).sents if len(i) > 0]

    def token(self, string: str):
        return [str(i) for i in self.nlp.tokenizer(string)]

    def keyword(self, string: str, n: int = 10):
        return self._get_keyword(self.nlp(string), string, n)

    def ner(self, string: str, n: int = None):
        keywords = self.nlp(string).ents
        return keywords[:min(len(keywords), n)] if n is not None else keywords

    @property
    def language(self):
        return self.nlp.lang



# coding=utf-8
# Copyright 2020 The HuggingFace Team All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A subclass of `Trainer` specific to Question-Answering tasks
"""

from transformers import Trainer, is_torch_tpu_available
from transformers.trainer_utils import PredictionOutput


if is_torch_tpu_available():
    import torch_xla.core.xla_model as xm
    import torch_xla.debug.metrics as met


class QuestionAnsweringTrainer(Trainer):
    def __init__(self, *args, eval_examples=None, post_process_function=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.eval_examples = eval_examples
        self.post_process_function = post_process_function

    def evaluate(self, eval_dataset=None, eval_examples=None, ignore_keys=None, metric_key_prefix: str = "eval"):
        eval_dataset = self.eval_dataset if eval_dataset is None else eval_dataset
        eval_dataloader = self.get_eval_dataloader(eval_dataset)
        eval_examples = self.eval_examples if eval_examples is None else eval_examples

        # Temporarily disable metric computation, we will do it in the loop here.
        compute_metrics = self.compute_metrics
        self.compute_metrics = None
        eval_loop = self.prediction_loop if self.args.use_legacy_prediction_loop else self.evaluation_loop
        try:
            output = eval_loop(
                eval_dataloader,
                description="Evaluation",
                # No point gathering the predictions if there are no metrics, otherwise we defer to
                # self.args.prediction_loss_only
                prediction_loss_only=True if compute_metrics is None else None,
                ignore_keys=ignore_keys,
            )
        finally:
            self.compute_metrics = compute_metrics

        if self.post_process_function is not None and self.compute_metrics is not None:
            eval_preds = self.post_process_function(eval_examples, eval_dataset, output.predictions)
            metrics = self.compute_metrics(eval_preds)

            # Prefix all keys with metric_key_prefix + '_'
            for key in list(metrics.keys()):
                if not key.startswith(f"{metric_key_prefix}_"):
                    metrics[f"{metric_key_prefix}_{key}"] = metrics.pop(key)

            self.log(metrics)
        else:
            metrics = {}

        if self.args.tpu_metrics_debug or self.args.debug:
            # tpu-comment: Logging debug metrics for PyTorch/XLA (compile, execute times, ops, etc.)
            xm.master_print(met.metrics_report())

        self.control = self.callback_handler.on_evaluate(self.args, self.state, self.control, metrics)
        return metrics

    def predict(self, predict_dataset, predict_examples, ignore_keys=None, metric_key_prefix: str = "test"):
        predict_dataloader = self.get_test_dataloader(predict_dataset)

        # Temporarily disable metric computation, we will do it in the loop here.
        compute_metrics = self.compute_metrics
        self.compute_metrics = None
        eval_loop = self.prediction_loop if self.args.use_legacy_prediction_loop else self.evaluation_loop
        try:
            output = eval_loop(
                predict_dataloader,
                description="Prediction",
                # No point gathering the predictions if there are no metrics, otherwise we defer to
                # self.args.prediction_loss_only
                prediction_loss_only=True if compute_metrics is None else None,
                ignore_keys=ignore_keys,
            )
        finally:
            self.compute_metrics = compute_metrics

        if self.post_process_function is None or self.compute_metrics is None:
            return output

        predictions = self.post_process_function(predict_examples, predict_dataset, output.predictions, "predict")
        metrics = self.compute_metrics(predictions)

        # Prefix all keys with metric_key_prefix + '_'
        for key in list(metrics.keys()):
            if not key.startswith(f"{metric_key_prefix}_"):
                metrics[f"{metric_key_prefix}_{key}"] = metrics.pop(key)

        return PredictionOutput(predictions=predictions.predictions, label_ids=predictions.label_ids, metrics=metrics)



#!/usr/bin/env python

# Python wrapper for METEOR implementation, by Xinlei Chen
# Acknowledge Michael Denkowski for the generous discussion and help
from __future__ import division

import atexit
import logging
import os
import re
import subprocess
import sys
import threading

import psutil
from ..text_normalization import text_normalization

import numpy as np


# Assumes meteor-1.5.jar is in the same directory as meteor.py.  Change as needed.
METEOR_JAR = 'meteor-1.5.jar'


def enc(s):
    return s.encode('utf-8')


def dec(s):
    return s.decode('utf-8')


class Meteor:

    def __init__(self, normalize_hypothesis=True):
        # Used to guarantee thread safety
        self.lock = threading.Lock()
        self.normalize_hypothesis = normalize_hypothesis

        mem = '2G'
        mem_available_G = psutil.virtual_memory().available / 1E9
        if mem_available_G < 2:
            logging.warning("There is less than 2GB of available memory.\n"
                            "Will try with limiting Meteor to 1GB of memory but this might cause issues.\n"
                            "If you have problems using Meteor, "
                            "then you can try to lower the `mem` variable in meteor.py")
            mem = '1G'

        meteor_cmd = ['java', '-jar', '-Xmx{}'.format(mem), METEOR_JAR,
                      '-', '-', '-stdio', '-l', 'en', '-norm']
        env = os.environ.copy()
        env['LC_ALL'] = "C"
        self.meteor_p = subprocess.Popen(meteor_cmd,
                                         cwd=os.path.dirname(os.path.abspath(__file__)),
                                         env=env,
                                         stdin=subprocess.PIPE,
                                         stdout=subprocess.PIPE,
                                         stderr=subprocess.PIPE)

        atexit.register(self.close)

    def close(self):
        with self.lock:
            if self.meteor_p:
                self.meteor_p.kill()
                self.meteor_p.wait()
                self.meteor_p = None
        # if the user calls close() manually, remove the
        # reference from atexit so the object can be garbage-collected.
        if atexit is not None and atexit.unregister is not None:
            atexit.unregister(self.close)

    def get_score(self, hyps, refs):
        scores = []
        eval_line = 'EVAL'
        with self.lock:
            assert len(hyps) == len(refs), f"{len(hyps)} != {len(refs)}"
            for h, r in zip(hyps, refs):
                h = h.encode() if type(h) is str else h
                r = r if type(r) is list else [r]
                r = [_r.encode() if type(_r) is str else _r for _r in r]
                if self.normalize_hypothesis:
                    h = text_normalization(h.decode()).encode('utf-8')
                stat = self._stat(h, r)
                eval_line += ' ||| {}'.format(stat)
            self.meteor_p.stdin.write(enc('{}\n'.format(eval_line)))
            self.meteor_p.stdin.flush()
            for i in range(0, len(hyps)):
                v = self.meteor_p.stdout.readline()
                try:
                    scores.append(float(dec(v.strip())))
                except:
                    sys.stderr.write("Error handling value: {}\n".format(v))
                    sys.stderr.write("Decoded value: {}\n".format(dec(v.strip())))
                    sys.stderr.write("eval_line: {}\n".format(eval_line))
                    # You can try uncommenting the next code line to show stderr from the Meteor JAR.
                    # If the Meteor JAR is not writing to stderr, then the line will just hang.
                    # sys.stderr.write("Error from Meteor:\n{}".format(self.meteor_p.stderr.read()))
                    raise
            score = float(dec(self.meteor_p.stdout.readline()).strip())

        return np.array(scores)

    def compute_score(self, gts, res):
        assert (gts.keys() == res.keys())
        imgIds = gts.keys()
        scores = []

        eval_line = 'EVAL'
        with self.lock:
            for i in imgIds:
                assert (len(res[i]) == 1)
                hypo = res[i][0]
                if self.normalize_hypothesis:
                    hypo = text_normalization(hypo.decode()).encode('utf-8')
                stat = self._stat(hypo, gts[i])
                eval_line += ' ||| {}'.format(stat)

            self.meteor_p.stdin.write(enc('{}\n'.format(eval_line)))
            self.meteor_p.stdin.flush()
            for _ in range(0, len(imgIds)):
                v = self.meteor_p.stdout.readline()
                try:
                    scores.append(float(dec(v.strip())))
                except:
                    sys.stderr.write("Error handling value: {}\n".format(v))
                    sys.stderr.write("Decoded value: {}\n".format(dec(v.strip())))
                    sys.stderr.write("eval_line: {}\n".format(eval_line))
                    # You can try uncommenting the next code line to show stderr from the Meteor JAR.
                    # If the Meteor JAR is not writing to stderr, then the line will just hang.
                    # sys.stderr.write("Error from Meteor:\n{}".format(self.meteor_p.stderr.read()))
                    raise
            score = float(dec(self.meteor_p.stdout.readline()).strip())

        return score, scores

    def method(self):
        return "METEOR"

    def _stat(self, hypothesis_str, reference_list):
        hypothesis_str = hypothesis_str.decode()
        reference_list = [ref.decode() for ref in reference_list]

        # SCORE ||| reference 1 words ||| reference n words ||| hypothesis words
        hypothesis_str = hypothesis_str.replace('|||', '')
        score_line = ' ||| '.join(('SCORE', ' ||| '.join(reference_list), hypothesis_str))
        score_line = re.sub(r'\s+', ' ', score_line)
        self.meteor_p.stdin.write(enc(score_line))
        self.meteor_p.stdin.write(enc('\n'))
        self.meteor_p.stdin.flush()
        return dec(self.meteor_p.stdout.readline()).strip()

    def _score(self, hypothesis_str, reference_list):
        with self.lock:
            # SCORE ||| reference 1 words ||| reference n words ||| hypothesis words
            hypothesis_str = hypothesis_str.replace('|||', '').replace('  ', ' ')
            score_line = ' ||| '.join(('SCORE', ' ||| '.join(reference_list), hypothesis_str))
            self.meteor_p.stdin.write(enc('{}\n'.format(score_line)))
            self.meteor_p.stdin.flush()
            stats = dec(self.meteor_p.stdout.readline()).strip()
            eval_line = 'EVAL ||| {}'.format(stats)
            # EVAL ||| stats
            self.meteor_p.stdin.write(enc('{}\n'.format(eval_line)))
            self.meteor_p.stdin.flush()
            score = float(dec(self.meteor_p.stdout.readline()).strip())
            # bug fix: there are two values returned by the jar file, one average, and one all, so do it twice
            # thanks for Andrej for pointing this out
            score = float(dec(self.meteor_p.stdout.readline()).strip())
        return score

    def __del__(self):
        self.close()


import glob
import json
import os
import logging
import string
import random
from os.path import join as pj
from typing import List
from itertools import product
from distutils.dir_util import copy_tree

from .trainer import Trainer, to_list
from .automatic_evaluation import evaluate
from .data import DEFAULT_CACHE_DIR


__all__ = 'GridSearcher'


def get_random_string(length: int = 6, exclude: List = None):
    tmp = ''.join(random.choice(string.ascii_lowercase) for _ in range(length))
    if exclude:
        while tmp in exclude:
            tmp = ''.join(random.choice(string.ascii_lowercase) for _ in range(length))
    return tmp


class GridSearcher:
    """ Grid search (epoch, batch, lr, random_seed, label_smoothing) """

    def __init__(self, checkpoint_dir: str, dataset_path: str = "asahi417/qg_squad", dataset_name: str = "default",
                 input_types: List or str = 'paragraph_answer', output_types: List or str = 'question',
                 prefix_types: List or str = 'qg', model: str = 't5-small', fp16: bool = False,
                 gradient_accumulation_steps: List or int = 4, metric: str = 'validation/Bleu_4',
                 epoch: int = 10, epoch_partial: int = 2, n_max_config: int = 5, max_length: int = 512,
                 max_length_eval: int = None, max_length_output: int = 32, max_length_output_eval: int = None,
                 prediction_aggregation: str = 'first', prediction_level: str = 'sentence',
                 batch: int = 128, batch_eval: int = 32, n_beams_eval: int = 4, lr: List or float = 1e-4,
                 label_smoothing: List or float = None, random_seed: List or int = 42, language: str = 'en',
                 normalize: bool = True, use_auth_token: bool = False):

        # evaluation configs
        max_length_eval = max_length if max_length_eval is None else max_length_eval
        max_length_output_eval = max_length_output if max_length_output_eval is None else max_length_output_eval
        self.eval_config = {
            'max_length_eval': max_length_eval, 'max_length_output_eval': max_length_output_eval,
            'n_beams_eval': n_beams_eval, 'prediction_aggregation': prediction_aggregation,
            'prediction_level': prediction_level, 'language': language, 'normalize': normalize
        }

        # static configs
        self.static_config = {
            'dataset_path': dataset_path, 'dataset_name': dataset_name, 'input_types': input_types,
            'output_types': output_types, 'model': model, 'fp16': fp16, 'batch': batch, 'epoch': epoch,
            'max_length': max_length, 'max_length_output': max_length_output, 'prefix_types': prefix_types
        }

        # dynamic config
        self.epoch = epoch
        self.epoch_partial = epoch_partial
        self.batch_eval = batch_eval
        self.checkpoint_dir = checkpoint_dir
        self.n_max_config = n_max_config
        self.use_auth_token = use_auth_token
        self.split, self.metric = metric.split('/')

        self.dynamic_config = {
            'lr': to_list(lr),
            'label_smoothing': to_list(label_smoothing),
            'random_seed': to_list(random_seed),
            'gradient_accumulation_steps': to_list(gradient_accumulation_steps),
        }

        self.all_dynamic_configs = list(product(
            self.dynamic_config['lr'],
            self.dynamic_config['label_smoothing'],
            self.dynamic_config['random_seed'],
            self.dynamic_config['gradient_accumulation_steps'],
        ))

    def initialize_searcher(self):
        os.makedirs(self.checkpoint_dir, exist_ok=True)
        path_to_config = pj(self.checkpoint_dir, 'config_static.json')
        if os.path.exists(path_to_config):
            with open(path_to_config) as f:
                tmp = json.load(f)
            tmp_v = [tmp[k] for k in sorted(tmp.keys())]
            static_tmp_v = [self.static_config[k] for k in sorted(tmp.keys())]
            assert tmp_v == static_tmp_v, f'{str(tmp_v)}\n not matched \n{str(static_tmp_v)}'
        path_to_d_config = pj(self.checkpoint_dir, 'config_dynamic.json')
        if os.path.exists(path_to_d_config):
            with open(path_to_d_config) as f:
                tmp = json.load(f)

            tmp_v = [tmp[k] for k in sorted(tmp.keys())]
            dynamic_tmp_v = [self.dynamic_config[k] for k in sorted(tmp.keys())]

            assert tmp_v == dynamic_tmp_v
        path_to_e_config = pj(self.checkpoint_dir, 'config_eval.json')
        if os.path.exists(path_to_e_config):
            with open(path_to_e_config) as f:
                tmp = json.load(f)
            tmp_v = [tmp[k] for k in sorted(tmp.keys())]
            eval_tmp_v = [self.eval_config[k] for k in sorted(tmp.keys())]
            assert tmp_v == eval_tmp_v, f'{str(tmp_v)}\n not matched \n{str(eval_tmp_v)}'

        with open(path_to_config, 'w') as f:
            json.dump(self.static_config, f)
        with open(path_to_d_config, 'w') as f:
            json.dump(self.dynamic_config, f)
        with open(path_to_e_config, 'w') as f:
            json.dump(self.eval_config, f)

        # add file handler
        logger = logging.getLogger()
        file_handler = logging.FileHandler(pj(self.checkpoint_dir, 'grid_search.log'))
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(logging.Formatter('%(asctime)s %(levelname)-8s %(message)s'))
        logger.addHandler(file_handler)
        logging.info(f'INITIALIZE GRID SEARCHER: {len(self.all_dynamic_configs)} configs to try')

    def get_evaluator(self, overwrite: bool):
        # configure evaluation data types
        input_types = to_list(self.static_config['input_types'], sorting=False)
        output_types = to_list(self.static_config['output_types'], sorting=False)
        assert len(input_types) == len(output_types), f"{len(input_types)} != {len(output_types)}"
        if self.static_config['prefix_types'] is None:
            prefix_types = [None] * len(input_types)
        else:
            prefix_types = to_list(self.static_config['prefix_types'], sorting=False)
        tmp = [(i, o, p) for i, o, p in zip(input_types, output_types, prefix_types)]
        if len(tmp) > 1:
            tmp = [(i, o, p) for i, o, p in tmp if o in ['question', 'questions_answers']]
        assert len(tmp) == 1, tmp
        i, o, p = tmp[0]
        prefix = pj(
            DEFAULT_CACHE_DIR,
            "encoded_feature",
            f"{self.static_config['dataset_path']}{'.' + self.static_config['dataset_name'] if self.static_config['dataset_name'] != 'default' else ''}"
            f"{self.static_config['model']}.{self.eval_config['max_length_eval']}.{self.eval_config['max_length_output_eval']}.{i}.{o}"
        )
        data_cache_paths = {split: f"{prefix}.{split}.{p}.pkl" for split in ['test', 'validation']}

        def get_evaluation(_checkpoint_dir_model):
            return evaluate(
                export_dir=pj(_checkpoint_dir_model, 'eval'),
                batch_size=self.batch_eval,
                n_beams=self.eval_config['n_beams_eval'],
                model=_checkpoint_dir_model,
                max_length=self.eval_config['max_length_eval'],
                overwrite=overwrite,
                max_length_output=self.eval_config['max_length_output_eval'],
                bleu_only=True,
                dataset_path=self.static_config['dataset_path'],
                dataset_name=self.static_config['dataset_name'],
                input_type=i,
                output_type=o,
                prediction_aggregation=self.eval_config['prediction_aggregation'],
                prediction_level=self.eval_config['prediction_level'],
                language=self.eval_config['language'],
                use_auth_token=self.use_auth_token,
                data_caches=data_cache_paths)
        return get_evaluation

    def run(self, interval: int = 25, overwrite: bool = False):

        self.initialize_searcher()

        # instantiate evaluator
        evaluator = self.get_evaluator(overwrite)
        ###########
        # 1st RUN #
        ###########
        checkpoints = []
        ckpt_exist = {}
        for trainer_config in glob.glob(pj(self.checkpoint_dir, 'model_*', 'trainer_config.json')):
            with open(trainer_config, 'r') as f:
                ckpt_exist[os.path.dirname(trainer_config)] = json.load(f)

        for n, dynamic_config in enumerate(self.all_dynamic_configs):
            logging.info(f'## 1st RUN: Configuration {n}/{len(self.all_dynamic_configs)} ##')
            config = self.static_config.copy()
            tmp_dynamic_config = {
                'lr': dynamic_config[0],
                'label_smoothing': dynamic_config[1],
                'random_seed': dynamic_config[2],
                'gradient_accumulation_steps': dynamic_config[3]
            }
            config.update(tmp_dynamic_config)
            ex_dynamic_config = [(k_, [v[k] for k in sorted(tmp_dynamic_config.keys())]) for k_, v in ckpt_exist.items()]
            tmp_dynamic_config = [tmp_dynamic_config[k] for k in sorted(tmp_dynamic_config.keys())]
            duplicated_ckpt = [k for k, v in ex_dynamic_config if v == tmp_dynamic_config]

            if len(duplicated_ckpt) == 1:
                logging.info(f'skip as the config exists at {duplicated_ckpt} \n{config}')
                checkpoint_dir = duplicated_ckpt[0]
            elif len(duplicated_ckpt) == 0:
                ckpt_name_exist = [os.path.basename(k).replace('model_', '') for k in ckpt_exist.keys()]
                ckpt_name_made = [os.path.basename(c).replace('model_', '') for c in checkpoints]
                model_ckpt = get_random_string(exclude=ckpt_name_exist + ckpt_name_made)
                checkpoint_dir = pj(self.checkpoint_dir, f'model_{model_ckpt}')
            else:
                raise ValueError(f'duplicated checkpoints are found: \n {duplicated_ckpt}')

            if not os.path.exists(pj(checkpoint_dir, f'epoch_{self.epoch_partial}')):
                trainer = Trainer(
                    checkpoint_dir=checkpoint_dir, disable_log=True, use_auth_token=self.use_auth_token, **config)
                trainer.train(
                    epoch_partial=self.epoch_partial, epoch_save=1, interval=interval)

            checkpoints.append(checkpoint_dir)

        metrics = {}
        for n, checkpoint_dir in enumerate(checkpoints):
            logging.info(f'## 1st RUN (EVAL): Configuration {n}/{len(checkpoints)} ##')
            checkpoint_dir_model = pj(checkpoint_dir, f'epoch_{self.epoch_partial}')
            try:
                metric = evaluator(checkpoint_dir_model)
                # except Exception:
                metrics[checkpoint_dir_model] = metric[self.split][self.metric]
            except Exception as e:
                logging.exception("Error while computing metric")

        metrics = sorted(metrics.items(), key=lambda x: x[1], reverse=True)
        with open(pj(self.checkpoint_dir, f"metric.{self.eval_config['prediction_level']}.1st.json"), 'w') as f:
            json.dump(metrics, f)

        logging.info(f'1st RUN RESULTS ({self.split}/{self.metric})')
        for n, (k, v) in enumerate(metrics):
            logging.info(f'\t * rank: {n} | metric: {round(v, 3)} | model: {k} |')

        if self.epoch_partial == self.epoch:
            logging.info('No 2nd phase as epoch_partial == epoch')
            return

        ###########
        # 2nd RUN #
        ###########
        metrics = metrics[:min(len(metrics), self.n_max_config)]
        checkpoints = []
        for n, (checkpoint_dir_model, _metric) in enumerate(metrics):
            logging.info(f'## 2nd RUN: Configuration {n}/{len(metrics)}: {self.split}/{self.metric} = {_metric}')
            model_ckpt = os.path.dirname(checkpoint_dir_model)
            if not os.path.exists(pj(model_ckpt, f'epoch_{self.epoch}')):
                trainer = Trainer(checkpoint_dir=model_ckpt, disable_log=True, use_auth_token=self.use_auth_token)
                trainer.train(epoch_save=1, interval=interval)

            checkpoints.append(model_ckpt)

        metrics = {}
        for n, checkpoint_dir in enumerate(checkpoints):
            logging.info(f'## 2nd RUN (EVAL): Configuration {n}/{len(checkpoints)} ##')
            for checkpoint_dir_model in sorted(glob.glob(pj(checkpoint_dir, 'epoch_*'))):
                if int(checkpoint_dir_model.split(os.path.sep)[-1].replace('epoch_', '')) > self.static_config['epoch']:
                    continue
                metric = evaluator(checkpoint_dir_model)
                metrics[checkpoint_dir_model] = metric[self.split][self.metric]

        metrics = sorted(metrics.items(), key=lambda x: x[1], reverse=True)
        logging.info(f'2nd RUN RESULTS: \n{str(metrics)}')
        for n, (k, v) in enumerate(metrics):
            logging.info(f'\t * rank: {n} | metric: {round(v, 3)} | model: {k} |')

        with open(pj(self.checkpoint_dir, f"metric.{self.eval_config['prediction_level']}.2nd.json"), 'w') as f:
            json.dump(metrics, f)

        best_model_ckpt, best_metric_score = metrics[0]
        epoch = int(best_model_ckpt.split(os.path.sep)[-1].replace('epoch_', ''))
        if epoch == self.static_config['epoch']:
            ###########
            # 3rd RUN #
            ###########
            best_model_dir = os.path.dirname(best_model_ckpt)
            logging.info(f'## 3rd RUN: target model: {best_model_dir} (metric: {best_metric_score}) ##')
            metric_list = [[epoch, best_metric_score]]
            with open(pj(best_model_dir, 'trainer_config.json')) as f:
                config = json.load(f)
            while True:
                epoch += 1
                logging.info(f'## 3rd RUN (TRAIN): epoch {epoch} ##')
                config['epoch'] = epoch
                with open(pj(best_model_dir, 'trainer_config.additional_training.json'), 'w') as f:
                    json.dump(config, f)
                checkpoint_dir_model = pj(best_model_dir, f'epoch_{epoch}')
                if not os.path.exists(checkpoint_dir_model):
                    trainer = Trainer(
                        checkpoint_dir=best_model_dir, config_file='trainer_config.additional_training.json',
                        disable_log=True, use_auth_token=self.use_auth_token)
                    trainer.train(epoch_save=1, interval=interval)
                logging.info(f'## 3rd RUN (EVAL): epoch {epoch} ##')
                metric = evaluator(checkpoint_dir_model)
                tmp_metric_score = metric[self.split][self.metric]
                metric_list.append([epoch, tmp_metric_score])
                logging.info(f'\t tmp metric: {tmp_metric_score}')
                if best_metric_score > tmp_metric_score:
                    logging.info('\t finish 3rd phase (no improvement)')
                    break
                else:
                    logging.info(f'\t tmp metric improved the best model ({best_metric_score} --> {tmp_metric_score})')
                    best_metric_score = tmp_metric_score
            logging.info(f'3rd RUN RESULTS: {best_model_dir}')
            for k, v in metric_list:
                logging.info(f'\t epoch {k}: {v}')

            metric_list = sorted(metric_list, key=lambda x: x[1], reverse=True)
            with open(pj(self.checkpoint_dir, f"metric.{self.eval_config['prediction_level']}.3rd.json"), 'w') as f:
                json.dump(metric_list, f)
            best_model_ckpt = pj(best_model_dir, f'epoch_{metric_list[0][0]}')

        copy_tree(best_model_ckpt, pj(self.checkpoint_dir, 'best_model'))



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    readme = f.read()
VERSION = '0.0.7'
NAME = 'lmqg'
LICENSE = 'MIT License'
setup(
    name=NAME,
    packages=find_packages(exclude=['tests', 'misc', 'asset']),
    version=VERSION,
    license=LICENSE,
    description='Language Model for Question Generation.',
    url='https://github.com/asahi417/lm-question-generation',
    download_url="https://github.com/asahi417/lm-question-generation/archive/v{}.tar.gz".format(VERSION),
    keywords=['language model', 'question-answering', 'question-generation'],
    long_description=readme,
    long_description_content_type="text/markdown",
    author='Asahi Ushio',
    author_email='asahi1992ushio@gmail.com',
    classifiers=[
        'Development Status :: 4 - Beta',       # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',      # Define that your audience are developers
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        f'License :: OSI Approved :: {LICENSE}',   # Again, pick a license
        'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
      ],
    include_package_data=True,
    test_suite='tests',
    extras_require={
        "api": [
            'uvicorn',
            'fastapi',
            'pydantic',
            'spacy_ke',
            'pydantic',
            'protobuf',
            'lmppl'
        ]
    },
    install_requires=[
        'psutil',
        'pytextrank',
        "torch",
        "tqdm",
        "requests",
        "pandas",
        "numpy",
        "transformers<=4.21.2",  # push-to-model is not working for latest version
        "huggingface-hub<=0.9.1",
        "sentencepiece",
        "datasets",
        "spacy",
        'sudachipy',
        'sudachidict_core',
        'bert-score',
        'pyemd',  # to compute moverscore
        'evaluate',
        "wandb",
        "ray",
        "ray[tune]"
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'lmqg-train-search = lmqg.lmqg_cl.model_finetuning:main_training_search',
            'lmqg-eval = lmqg.lmqg_cl.model_evaluation:main',
            'lmqg-eval-qag = lmqg.lmqg_cl.model_evaluation_qag:main',
            'lmqg-eval-qa = lmqg.lmqg_cl.model_evaluation_qa:main',
            'lmqg-push-to-hf = lmqg.lmqg_cl.push_to_hf:main',
            'lmqg-generate-qa = lmqg.lmqg_cl.model_evaluation_qa_based_metric:main_generate_qa_pair',
            'lmqg-qae = lmqg.lmqg_cl.model_evaluation_qa_based_metric:main_qa_model_training'
        ]
    }
)




""" Apply same preprocessing to the question as the gold reference of NQG.
https://github.com/xinyadu/nqg/blob/master/data/processed/tgt-test.txt
"""
import re

_tok_dict = {"(": "-lrb-", ")": "-rrb-", "[": "-lsb-", "]": "-rsb-", "{": "-lcb-", "}": "-rcb-"}


def text_normalization(text):

    # lower casing
    text = text.lower()

    # convert into the representation of the gold question
    for k, v in _tok_dict.items():
        text = text.replace(k, v)

    # add half space before eg) What? --> What ?
    for s in ["'s", "?", ",", "'", "$", "%"]:
        text = re.sub(r'\s*\{}'.format(s), ' {}'.format(s), text)

    # quote is `` '', not "" ""
    text = re.sub(r'"([^"]+)"', r"`` \1 ''", text)

    # more than 2 continuation, add half space eg) A... --> A ...
    for s in [".", "-"]:
        text = re.sub(r'\s*([{}].{{2}})'.format(s), r' \1', text)

    return text



#!/usr/bin/env python
# 
# File Name : bleu.py
#
# Description : Wrapper for BLEU scorer.
#
# Creation Date : 06-01-2015
# Last Modified : Thu 19 Mar 2015 09:13:28 PM PDT
# Authors : Hao Fang <hfang@uw.edu> and Tsung-Yi Lin <tl483@cornell.edu>
import numpy as np

from .bleu_scorer import BleuScorer
from ..text_normalization import text_normalization


class Bleu:
    def __init__(self, n=4, normalize_hypothesis=True):
        self._n = n
        self._normalize_hypothesis = normalize_hypothesis
        self._hypo_for_image = {}
        self.ref_for_image = {}

    def compute_score(self, gts, res):

        assert(gts.keys() == res.keys())
        bleu_scorer = BleuScorer(n=self._n)
        for _id in gts.keys():
            hypo = res[_id]
            ref = gts[_id]

            # Sanity check.
            assert(type(hypo) is list)
            assert(len(hypo) == 1)
            assert(type(ref) is list)
            assert(len(ref) >= 1)
            hypo = hypo[0]
            if self._normalize_hypothesis:
                if type(hypo) == str:
                    hypo = hypo.encode()
                assert type(hypo) == bytes, f"{hypo} ({type(hypo)})"
                hypo = text_normalization(hypo.decode()).encode('utf-8')
            bleu_scorer += (hypo, ref)

        score, scores = bleu_scorer.compute_score(option='closest', verbose=0)

        return score, scores

    def get_score(self, hyps, refs):
        assert len(hyps) == len(refs), f"{len(hyps)} != {len(refs)}"
        bleu_scorer = BleuScorer(n=self._n)
        for h, r in zip(hyps, refs):
            assert type(h) is str, h
            r = r if type(r) is list else [r]
            r = [_r.encode() if type(_r) is str else _r for _r in r]
            h = h.encode() if type(h) is str else h
            if self._normalize_hypothesis:
                assert type(h) == bytes, f"{h} ({type(h)})"
                h = text_normalization(h.decode()).encode('utf-8')
            bleu_scorer += (h, r)
        _, scores = bleu_scorer.compute_score(option='closest', verbose=0)
        return np.array(scores)

    @staticmethod
    def method():
        return "Bleu"



from .text_normalization import text_normalization
from .bleu.bleu import Bleu
from .meteor.meteor import Meteor
from .rouge import Rouge
from .bertscore import BERTScore
from .moverscore import MoverScore
from .qa_aligned_f1_score import QAAlignedF1Score



""" Evaluate Multitask QAG Model with QAG metric (eg. QA Aligned F1). """
import json
import logging
import argparse
import os
from itertools import chain
from statistics import mean

from datasets import load_dataset
from lmqg import TransformersQG
from lmqg.automatic_evaluation_tool import QAAlignedF1Score, Bleu, Meteor, Rouge, BERTScore, MoverScore
from lmqg.spacy_module import SpacyPipeline
from lmqg.automatic_evaluation import LANG_NEED_TOKENIZATION


logging.basicConfig(format='%(asctime)s %(levelname)-8s %(message)s', level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')


def get_options():
    parser = argparse.ArgumentParser(description='QAG evaluation.')
    parser.add_argument('-m', '--model', default=None, type=str)
    parser.add_argument('--model-ae', default=None, type=str)
    parser.add_argument('--max-length', default=512, type=int, help='')
    parser.add_argument('--max-length-output', default=256, type=int, help='')
    parser.add_argument('-d', '--dataset-path', help='huggingface datasets alias', default='lmqg/qg_squad', type=str)
    parser.add_argument('--dataset-name', help='huggingface datasets name', default='default', type=str)
    parser.add_argument('--test-split', help='the name of test split', default='test', type=str)
    parser.add_argument('--validation-split', help='the name of validation split', default='validation', type=str)
    parser.add_argument('--n-beams', default=4, type=int)
    parser.add_argument('--batch-size', default=16, type=int)
    parser.add_argument('-l', '--language', help='', default='en', type=str)
    parser.add_argument('--use-auth-token', help='', action='store_true')
    parser.add_argument('-e', '--export-dir', required=True, type=str)
    parser.add_argument('--hyp-test', default=None, type=str)
    parser.add_argument('--hyp-dev', default=None, type=str)
    parser.add_argument('--overwrite-prediction', help='', action='store_true')
    parser.add_argument('--overwrite-metric', help='', action='store_true')
    parser.add_argument('--use-reference-answer', action='store_true')
    return parser.parse_args()


def main():
    opt = get_options()
    os.makedirs(opt.export_dir, exist_ok=True)
    metrics = [
        (QAAlignedF1Score(target_metric='f1', base_metric='bertscore', language=opt.language),
         "QAAlignedF1Score (BERTScore)"),
        (QAAlignedF1Score(target_metric='recall', base_metric='bertscore', language=opt.language),
         "QAAlignedRecall (BERTScore)"),
        (QAAlignedF1Score(target_metric='precision', base_metric='bertscore', language=opt.language),
         "QAAlignedPrecision (BERTScore)"),
        (QAAlignedF1Score(target_metric='f1', base_metric='moverscore', language=opt.language),
         "QAAlignedF1Score (MoverScore)"),
        (QAAlignedF1Score(target_metric='recall', base_metric='moverscore', language=opt.language),
         "QAAlignedRecall (MoverScore)"),
        (QAAlignedF1Score(target_metric='precision', base_metric='moverscore', language=opt.language),
         "QAAlignedPrecision (MoverScore)"),
        (Bleu(4), ["Bleu_1", "Bleu_2", "Bleu_3", "Bleu_4"]),
        (Meteor(), "METEOR"),
        (Rouge(), "ROUGE_L"),
        (BERTScore(language=opt.language), "BERTScore"),
        (MoverScore(language=opt.language), 'MoverScore')
    ]

    def load_model():
        if opt.model is not None:
            _model = TransformersQG(opt.model,
                                    model_ae=opt.model_ae,
                                    skip_overflow_error=True,
                                    drop_answer_error_text=True,
                                    language=opt.language,
                                    max_length=opt.max_length,
                                    max_length_output=opt.max_length_output,
                                    use_auth_token=opt.use_auth_token)
            _model.eval()
            return _model
        raise ValueError(f"require `-m` or `--model`")

    if opt.model_ae is not None:
        metric_file = f"{opt.export_dir}/metric.first.answer.paragraph.questions_answers." \
                      f"{opt.dataset_path.replace('/', '_')}.{opt.dataset_name}." \
                      f"{opt.model_ae.replace('/', '_')}.json"
    else:
        metric_file = f"{opt.export_dir}/metric.first.answer.paragraph.questions_answers." \
                      f"{opt.dataset_path.replace('/', '_')}.{opt.dataset_name}.json"
    if os.path.exists(metric_file):
        with open(metric_file) as f:
            output = json.load(f)
    else:
        output = {}
    spacy_model = SpacyPipeline(language=opt.language) if opt.language in LANG_NEED_TOKENIZATION else None
    for _split, _file in zip([opt.test_split, opt.validation_split], [opt.hyp_test, opt.hyp_dev]):
        if _file is None:
            if opt.model_ae is not None:
                _file = f"{opt.export_dir}/samples.{_split}.hyp.paragraph.questions_answers." \
                        f"{opt.dataset_path.replace('/', '_')}.{opt.dataset_name}." \
                        f"{opt.model_ae.replace('/', '_')}.txt"
            else:
                _file = f"{opt.export_dir}/samples.{_split}.hyp.paragraph.questions_answers." \
                        f"{opt.dataset_path.replace('/', '_')}.{opt.dataset_name}.txt"

        logging.info(f'generate qa for split {_split}')
        if _split not in output:
            output[_split] = {}

        dataset = load_dataset(opt.dataset_path, None if opt.dataset_name == 'default' else opt.dataset_name,
                               split=_split, use_auth_token=opt.use_auth_token)
        df = dataset.to_pandas()

        # formatting data into qag format
        model_input = []
        gold_reference = []
        model_highlight = []
        for paragraph, g in df.groupby("paragraph"):
            model_input.append(paragraph)
            model_highlight.append(g['answer'].tolist())
            gold_reference.append(' | '.join([
                f"question: {i['question']}, answer: {i['answer']}" for _, i in g.iterrows()
            ]))
        prediction = None
        if not opt.overwrite_prediction and os.path.exists(_file):
            with open(_file) as f:
                _prediction = f.read().split('\n')
            if len(_prediction) != len(gold_reference):
                logging.warning(f"found prediction file at {_file} but length not match "
                                f"({len(_prediction)} != {len(gold_reference)})")
            else:
                prediction = _prediction
        if prediction is None:
            model = load_model()
            # model prediction
            if not opt.use_reference_answer:
                logging.info("model prediction: (qag model)")
                prediction = model.generate_qa(
                    list_context=model_input,
                    num_beams=opt.n_beams,
                    batch_size=opt.batch_size)
            else:
                logging.info("model prediction: (qg model, answer fixed by reference)")
                model_input_flat = list(chain(*[[i] * len(h) for i, h in zip(model_input, model_highlight)]))
                model_highlight_flat = list(chain(*model_highlight))
                prediction_flat = model.generate_q(
                    list_context=model_input_flat,
                    list_answer=model_highlight_flat,
                    num_beams=opt.n_beams,
                    batch_size=opt.batch_size)
                _index = 0
                prediction = []
                for h in model_highlight:
                    questions = prediction_flat[_index:_index+len(h)]
                    answers = model_highlight_flat[_index:_index+len(h)]
                    prediction.append(list(zip(questions, answers)))
                    _index += len(h)

            # formatting prediction
            prediction = [' | '.join([f"question: {q}, answer: {a}" for q, a in p]) if p is not None else "" for p in prediction]
            assert len(prediction) == len(model_input), f"{len(prediction)} != {len(model_input)}"
            with open(_file, 'w') as f:
                f.write('\n'.join(prediction))

        for metric, metric_name in metrics:
            metric_name_list = [metric_name] if type(metric_name) is str else metric_name
            if opt.overwrite_metric or any(m not in output[_split] for m in metric_name_list):
                if spacy_model is not None and (type(metric_name) is list or not metric_name.startswith("QAAligned")):
                    prediction = [' '.join(spacy_model.token(i)) for i in prediction]
                scores = metric.get_score(prediction, gold_reference)
                if type(metric_name) is list:
                    for _metric_name, _score in zip(metric_name, scores):
                        output[_split][_metric_name] = mean(_score.tolist())
                else:
                    output[_split][metric_name] = mean(scores.tolist())

    with open(metric_file, "w") as f:
        json.dump(output, f)




""" Parameter optimization of Question Generation Finetuning. """
import argparse
import logging

from lmqg import GridSearcher, Trainer


def arguments(parser):
    parser.add_argument('-c', '--checkpoint-dir', help='directory to save checkpoint', required=True, type=str)
    parser.add_argument('-d', '--dataset-path', help='huggingface datasets alias', default='lmqg/qg_squad', type=str)
    parser.add_argument('--dataset-name', help='huggingface datasets name', default='default', type=str)
    parser.add_argument('-i', '--input-types',
                        help='input to the model (`paragraph_answer`/`sentence_answer`/`paragraph_sentence`)',
                        nargs='+', default='paragraph_answer', type=str)
    parser.add_argument('-o', '--output-types',
                        help='output of the model (`answer`/`question`)',
                        nargs='+', default='question', type=str)
    parser.add_argument('-p', '--prefix-types',
                        help='prefix type (`qg`/`ae`) which should be specified for T5 but not for mT5, BART, or mBART',
                        nargs='+', default=None, type=str)
    parser.add_argument('-m', '--model', help='pretrained language model', default='facebook/bart-large', type=str)
    parser.add_argument('-e', '--epoch', help='epoch', default=8, type=int)
    parser.add_argument('-b', '--batch', help='batch size', default=128, type=int)
    parser.add_argument('--fp16', help='fp16', action='store_true')
    parser.add_argument('--use-auth-token', help='', action='store_true')
    parser.add_argument('--max-length', default=512, type=int, help='max sequence length for input sequence')
    parser.add_argument('--max-length-output', default=32, type=int, help='max sequence length for output sequence')
    parser.add_argument('--interval', default=50, type=int)
    return parser


def arguments_training(parser):
    # training config
    parser.add_argument('--random-seed', help='random seed', default=1234, type=int)
    parser.add_argument('--label-smoothing', help='label smoothing', default=0.3, type=float)
    parser.add_argument('-l', '--lr', help='learning rate', default=1e-4, type=float)
    parser.add_argument('-g', '--gradient-accumulation-steps', help='number of gradient accumulation', default=1, type=int)
    # monitoring parameter
    parser.add_argument('--epoch-save', default=None, type=int)
    return parser


def arguments_training_search(parser):
    parser.add_argument('--n-beams-eval', help='number of beam at evaluation', default=4, type=int)
    parser.add_argument('--batch-eval', help='batch size at evaluation', default=8, type=int)
    parser.add_argument('--prediction-aggregation', help='`first`/`short`/`last`/`long`/`middle`', default='first', type=str)
    parser.add_argument('--prediction-level', help="`sentence`/`paragraph`/`answer`", default='answer', type=str)
    parser.add_argument('--metric', help='evaluation metric for grid search', default='validation/Bleu_4', type=str)
    parser.add_argument('--n-max-config', help='number of maximum configuration to select for the second stage',
                        default=5, type=int)
    parser.add_argument('--epoch-partial', help='epoch', default=2, type=int)
    parser.add_argument('--max-length-eval', help='max sequence length for input sequence at evaluation',
                        default=512, type=int)
    parser.add_argument('--max-length-output-eval', help='max sequence length for output sequence at evaluation',
                        default=64, type=int)
    parser.add_argument('-l', '--lr', help='learning rate', nargs='+', default=[1e-05, 5e-05, 1e-04], type=float)
    parser.add_argument('--label-smoothing', help='label smoothing', nargs='+', default=[0.0, 0.15], type=float)
    parser.add_argument('--random-seed', help='random seed', nargs='+', default=[1], type=int)
    parser.add_argument('-g', '--gradient-accumulation-steps', help='number of gradient accumulation',
                        nargs='+',  default=[1, 2, 4, 8], type=int)
    parser.add_argument('--overwrite', help='overwrite metric already computed', action='store_true')
    parser.add_argument('--language', help='language needs to be specified for evaluation mainly',
                        default='en', type=str)
    return parser


def main_training():
    parser = argparse.ArgumentParser(description='Fine-tuning on QG.')
    parser = arguments(parser)
    parser = arguments_training(parser)
    opt = parser.parse_args()

    logging.basicConfig(format='%(asctime)s %(levelname)-8s %(message)s', level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')
    # train model
    trainer = Trainer(
        checkpoint_dir=opt.checkpoint_dir,
        dataset_path=opt.dataset_path,
        dataset_name=opt.dataset_name,
        input_types=opt.input_types,
        output_types=opt.output_types,
        prefix_types=opt.prefix_types if opt.prefix_types is not None else None,
        random_seed=opt.random_seed,
        model=opt.model,
        epoch=opt.epoch,
        lr=opt.lr,
        batch=opt.batch,
        max_length=opt.max_length,
        max_length_output=opt.max_length_output,
        fp16=opt.fp16,
        gradient_accumulation_steps=opt.gradient_accumulation_steps,
        label_smoothing=opt.label_smoothing,
        use_auth_token=opt.use_auth_token)
    trainer.train(interval=opt.interval, epoch_save=opt.epoch_save)


def main_training_search():
    parser = argparse.ArgumentParser(description='Fine-tuning on QG with Grid Search.')
    parser = arguments(parser)
    parser = arguments_training_search(parser)
    opt = parser.parse_args()

    logging.basicConfig(format='%(asctime)s %(levelname)-8s %(message)s', level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')
    # train model
    trainer = GridSearcher(
        checkpoint_dir=opt.checkpoint_dir,
        dataset_path=opt.dataset_path,
        dataset_name=opt.dataset_name,
        input_types=opt.input_types,
        output_types=opt.output_types,
        prefix_types=opt.prefix_types if opt.prefix_types is not None else None,
        model=opt.model,
        fp16=opt.fp16,
        epoch=opt.epoch,
        metric=opt.metric,
        epoch_partial=opt.epoch_partial,
        batch=opt.batch,
        max_length=opt.max_length,
        max_length_output=opt.max_length_output,
        n_max_config=opt.n_max_config,
        gradient_accumulation_steps=opt.gradient_accumulation_steps,
        lr=opt.lr,
        label_smoothing=opt.label_smoothing,
        random_seed=opt.random_seed,
        batch_eval=opt.batch_eval,
        n_beams_eval=opt.n_beams_eval,
        prediction_aggregation=opt.prediction_aggregation,
        prediction_level=opt.prediction_level,
        max_length_output_eval=opt.max_length_output_eval,
        max_length_eval=opt.max_length_eval,
        language=opt.language,
        use_auth_token=opt.use_auth_token
    )
    trainer.run(interval=opt.interval, overwrite=opt.overwrite)




""" Push Models to Modelhub"""
import os
import argparse
import logging
import shutil
from distutils.dir_util import copy_tree
from os.path import join as pj
from huggingface_hub import create_repo

import transformers
from lmqg.lmqg_cl.readme_template import get_readme

logging.basicConfig(format='%(asctime)s %(levelname)-8s %(message)s', level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')


def main():
    parser = argparse.ArgumentParser(description='Push to Model hub')
    parser.add_argument('-m', '--model-checkpoint', required=True, type=str)
    parser.add_argument('-a', '--model-alias', required=True, type=str)
    parser.add_argument('-o', '--organization', required=True, type=str)
    parser.add_argument('--use-auth-token', help='Huggingface transformers argument of `use_auth_token`',
                        action='store_true')
    parser.add_argument('--skip-model-upload', help='', action='store_true')
    opt = parser.parse_args()

    assert os.path.exists(pj(opt.model_checkpoint, "pytorch_model.bin")), pj(opt.model_checkpoint, "pytorch_model.bin")
    logging.info(f"Upload {opt.model_checkpoint} to {opt.organization}/{opt.model_alias}")

    url = create_repo(opt.model_alias, organization=opt.organization, exist_ok=True)

    if not opt.skip_model_upload:
        tokenizer = transformers.AutoTokenizer.from_pretrained(opt.model_checkpoint, local_files_only=True)
        config = transformers.AutoConfig.from_pretrained(opt.model_checkpoint, local_files_only=True)
        if config.model_type == 't5':  # T5 model requires T5ForConditionalGeneration class
            model_class = transformers.T5ForConditionalGeneration.from_pretrained
        elif config.model_type == 'mt5':
            model_class = transformers.MT5ForConditionalGeneration.from_pretrained
        elif config.model_type == 'bart':
            model_class = transformers.BartForConditionalGeneration.from_pretrained
        elif config.model_type == 'mbart':
            model_class = transformers.MBartForConditionalGeneration.from_pretrained
        else:
            raise ValueError(f'unsupported model type: {config.model_type}')

        model = model_class(opt.model_checkpoint, config=config, local_files_only=True)
        args = {"use_auth_token": opt.use_auth_token, "repo_url": url, "organization": opt.organization}
        model.push_to_hub(opt.model_alias, **args)
        tokenizer.push_to_hub(opt.model_alias, **args)
        config.push_to_hub(opt.model_alias, **args)

    # upload remaining files
    copy_tree(f"{opt.model_checkpoint}", f"{opt.model_alias}")

    # config
    readme = get_readme(
        model_name=f"{opt.organization}/{opt.model_alias}",
        model_checkpoint=opt.model_checkpoint
    )
    with open(pj(opt.model_checkpoint, "README.md"), 'w') as f:
        f.write(readme)

    # upload remaining files
    copy_tree(f"{opt.model_checkpoint}", f"{opt.model_alias}")
    os.system(
        f"cd {opt.model_alias} && git lfs install && git add . && git commit -m 'model update' && git push && cd ../")
    shutil.rmtree(f"{opt.model_alias}")  # clean up the cloned repo
